#!/bin/ksh93

# License: CDDL 1.0
# Copyright (C) 2013 by Jens Elkner . All rights reserved.

export LC_CTYPE=C LC_NUMERIC=C LC_MESSAGES=C LC_TIME=C LC_COLLATE=C LC_MEASUREMENT=C

function getKey {
	typeset X=${1##*/}
	X=${X%%.*}
	X=${X// /_}
	print ${X//:}
}

function getRWaverage {
	typeset -F1 VAL MIN=$2 MAX=$2 SUM=$2
	typeset -a VALS=( $@ )
	for (( I=$# - 2; I > 1; I-- )); do
		VAL=${VALS[$I]}
		if (( $VAL < $MIN )); then
			MIN=$VAL
		elif (( $VAL > $MAX )); then
			MAX=$VAL
		fi
		(( SUM+=$VAL ))
	done
	if (( ${#VALS[*]} > 4 )); then
		(( SUM-=$MIN ))
		(( SUM-=$MAX ))
		(( SUM/=$# - 4 ))
	else
		(( SUM/=$# - 2 ))
	fi
	print $SUM
}

function getFBVals {
	typeset -a -F1 VALS MIN=( ) MAX=( ) SUM=( )
	typeset LINE LAST TNAME=$1		# filebench workload name
	integer I COUNT=0 DIFF=0
	while read LINE ; do
		[[ -z "$LINE" ]] && break			# no more results
		VALS=( ${LINE//\// } )
		if [[ ${TNAME} == 'oltp' ]] && (( VALS[4] < 1.0 )); then
			# on Linux sometimes threads go crazy and produce unusable results
			LAST="$LINE"
			continue
		fi
		(( COUNT++ ))
		if (( ${#SUM[*]} == 0 )); then
			MIN=( ${VALS[*]} )				# first res seen - init vals
			MAX=( ${VALS[*]} )
			SUM=( ${VALS[*]} )
			continue
		elif (( ${VALS[4]} < ${MIN[4]} )); then		# adjust min wrt. mb/s
			MIN=( ${VALS[*]} )
			(( DIFF++ ))
		elif (( ${VALS[4]} > ${MAX[4]} )); then		# adjust max wrt. mb/s
			MAX=( ${VALS[*]} )
			(( DIFF++ ))
		fi
		for (( I=${#VALS[*]} - 1; I >= 0; I-- )); do
			(( SUM[$I]+=${VALS[$I]} ))
		done
	done
	if (( COUNT == 0 )) && [[ ${TNAME} == 'oltp' && -n ${LAST} ]]; then
		# if all test gave crazy results (possibly on Linux) show it, so that
		# at least one sees, that the tests were made
		VALS=( ${LAST//\// } )
		MIN=( ${VALS[*]} )
		MAX=( ${VALS[*]} )
		SUM=( ${VALS[*]} )
		(( COUNT++ ))
	fi
	(( COUNT == 0 )) && return
	# since we look at mb/s only, we can drop max and min only if there were
	# at least 2 different records, otherwise it could screw up the average even
	# more
	if (( COUNT > 2 && DIFF > 0 )); then
		(( COUNT-=2 ))
		for (( I=${#SUM[*]} - 1; I >= 0; I-- )); do
			(( SUM[$I]-=${MIN[$I]} ))
			(( SUM[$I]-=${MAX[$I]} ))
			(( SUM[$I]/=$COUNT ))
		done
	else
		for (( I=${#SUM[*]} - 1; I >= 0; I-- )); do
			(( SUM[$I]/=$COUNT ))
		done
	fi
	print -- "${SUM[@]}"
}

function analyze {
	integer N=0
	typeset LINE KEY=${ getKey $1 ; }				# basically file basename
	typeset -F1 RV RW
	while read LINE; do
		[[ ${LINE:0:3} == 'TBD' ]] && break			# End of test res
		[[ ${LINE:0:5} != 'write' ]] && continue	# fast-forward to ddTest res
		(( N++ ))
		RV=${ getRWaverage $LINE ; }				# mkfile write
		read LINE
		WV=${ getRWaverage $LINE ; }				# dd read
		RES[${KEY}':RW:F'$N]="$RV $WV"
		(( TNAMES['RW']++ ))
		while read LINE; do
			[[ ${LINE:0:2} == '--' ]] && break		# heading
			[[ ${LINE:0:3} == 'TBD' ]] && break 2	# end of res
			[[ ${LINE:0:9} == ~(Ei)(filebench|==) ]] && continue	# heading
			if [[ ${LINE: -1:1} == ':' ]]; then
				LINE=${LINE%:}						# fb testname
				RES[${KEY}':'${LINE}':F'$N]=${ getFBVals ${LINE} ; }	# n res
				(( TNAMES[$LINE]++ ))
			fi
		done
	done < $1
	(( ${TESTS} < $N )) && TESTS=$N
}

function printHTMLProlog {
	print '<?xml version="1.0" encoding="ISO-8859-1"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<title>Filebench results</title>
<style type="text/css">
	caption {
		caption-side: top;
	}
	caption, .caption {
		text-align: left;
		font-weight: bold;
		color: blue;
	}
	.caption {
		padding-left: 0;
	}
	table {
		margin-left: 2em;
	}
	th {
		text-align: left;
		padding-left: 1em;
	}
	thead th {
		text-align: center;
		padding: 0 1ex 0 1ex;
	}
	td {
		padding: 0 1ex 0 1ex;
		font: monospace;
		text-align: right;
	}
</style>
</head>
<body>
<h1>HDD benchmark results</h1>
<p>Results of simple benchmarks using <b>mkfile</b>(1M), <b>dd</b>(1) and <b><a href="http://sourceforge.net/apps/mediawiki/filebench/index.php?title=Filebench">filebench</a></b>. The RW test is used to get a feeling for max. performance wrt. to sequential reads and writes. It writes out 4x32GiB (so ARC should be poissioned enough) using <b>mkfile</b>(1M) and after that it reads in those files using <b>dd</b>(1) in chunks of 128KiB. All other tests are made using the workflow definitions coming with filebench (fb) with <i>run 60</i>. The only parameter changed in the fb tests is the directory to point to the appropriate directory in the tested ZFS pool. We avoided explicitly changing other parameters to have "something comparable" out of the box. So depending on the workload/test, the results may <i>not</i> necessarily represent the max. performance the system + storage is able to provide! So to properly interprete the results, it is recommended to have a look at the workflow definition in question and related <a href="http://filebench.sourceforge.net/wiki/index.php/Workload_Model_Language">semantics</a>.</p>
<p>For more test details, have a look at <a href="./">*.txt</a> files and related sub directories.</p>
<p>NOTE: On Linux/Ubuntu we had to remove <b>useism</b> (use local instead of shared memory) from the <b>oltp.f</b> workload to get it work. Otherwise threads dead lock with FUTEX_WAIT when trying to attach the shared memory segment (ipc.c:ipc_ismmalloc).</p>
'
}

function printHTMLEpilog {
	print '
</body></html>'
}

function printResults {
	typeset VAL FILE NAMES 
	typeset FMT_TITLE FMT_THEAD FMT_TFOOT FMT_VALS FMT_EOS
	typeset FMT_TITLE2 FMT_THEAD2 FMT_TFOOT2 FMT_VALS2 FMT_EOS2
	integer I HTML=0 C
	typeset -a -F1 VALS
	[[ -n "$1" ]] && HTML=$1
	if (( $HTML > 1 )); then
		FMT_TITLE='<h2>%s</h2>\n<table>\n<thead>\n<tr><th>config</th><th>ops</th><th>ops/s</th><th>reads</th><th>writes</th><th>mb/s</th><th>CPU &micro;s/op</th><th>latency [ms]</th></tr>\n</thead><tbody>\n'
		FMT_THEAD='<tr><th colspan="8" class="caption">fb&nbsp;%d</th></tr>\n'
		FMT_ROW='<tr><th>%s</th><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>\n'
		FMT_TFOOT=''
		FMT_EOS='</tbody></table>\n'
		# RW
		FMT_TITLE2='<h2>%s</h2>\n<table>\n<thead>\n<tr><th>config</th><th>write [MiB/s]</th><th>read [MiB/s]</th></tr>\n</thead><tbody>\n'
		FMT_THEAD2='<tr><th colspan="3" class="caption">fb&nbsp;%d</th></tr>\n'
		FMT_ROW2='<tr><th>%s</th><td>%s</td><td>%s</td></tr>\n'
		FMT_TFOOT2=$FMT_TFOOT
		FMT_EOS2=$FMT_EOS
		printHTMLProlog
	elif (( $HTML )); then
		FMT_TITLE='<h2>%s</h2>\n'
		FMT_THEAD='<table><caption>fb&nbsp;%d</caption>\n<tbody>\n'
		FMT_ROW='<tr><th>%s</th><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>\n'
		FMT_TFOOT='</tbody></table>\n'
		FMT_EOS=''
		printHTMLProlog
		FMT_TITLE2=$FMT_TITLE2 FMT_THEAD2=$FMT_THEAD
		FMT_ROW2='<tr><th>%s</th><td>%s</td><td>%s</td></tr>\n'
		FMT_TFOOT2=$FMT_TFOOT FMT_EOS2=$FMT_EOS
	else
		FMT_TITLE='%s:\n'
		FMT_THEAD='\tfb %d\n'
		FMT_ROW='\t\t%24s:\t%12s\t%12s\t%12s\t%12s\t%8s\t%8s\t%8s\n'
		FMT_TFOOT='\n'
		FMT_EOS=''
		FMT_TITLE2=$FMT_TITLE2 FMT_THEAD2=$FMT_THEAD
		FMT_ROW2='\t\t%24s:\t%8s\t%8s\n'
		FMT_TFOOT2=$FMT_TFOOT FMT_EOS2=$FMT_EOS
	fi
	printf "$FMT_TITLE2" 'RW'
	for (( I=1; I <= $TESTS; I++ )); do
		printf "$FMT_THEAD2" $I
		for FILE in $FILES; do
			KEY=${ getKey $FILE ; }
			VAL=${RES[${KEY}':RW:F'${I}]}
			[[ -z $VAL ]] && continue
			VALS=( $VAL )
			printf "$FMT_ROW2" "$KEY" ${VALS[@]}
		done
		printf "$FMT_TFOOT2"
	done
	printf "$FMT_EOS2"
	NAMES=${ print "${!TNAMES[*]}" | sort ; }
	for NAME in $NAMES ; do
		[[ $NAME == 'RW' ]] && continue
		printf "$FMT_TITLE" "${NAME}"
		for (( I=1; I <= $TESTS; I++ )); do
			printf "$FMT_THEAD" $I
			C=0
			for FILE in $FILES; do
				KEY=${ getKey $FILE ; }
				VAL=${RES[${KEY}':'${NAME}':F'${I}]}
				[[ -z $VAL ]] && continue
				VALS=( $VAL )
				printf "$FMT_ROW" "$KEY" ${VALS[@]}
				(( C++ ))
			done
			(( $HTML != 0 && $C == 0 )) && print '<tr><th></th></tr>'
			printf "$FMT_TFOOT"
		done
		printf "$FMT_EOS"
	done
	(( $HTML )) && printHTMLEpilog
}

function showUsage {
	print "Usage: ${.sh.file##*/} [-F] [-T fname[,...]] [-f fmt] file ...

  file  .. output of doFilebench.sh -c ...
  fmt   .. 0 for print results in plain/txt (default), or 1 for compact html ,
           or 2 for html with table headings
  fname .. name of the function to trace
"
}

typeset -A RES=( )
typeset -i TESTS=0 FMT=0
typeset -A -i TNAMES=( )
typeset FILES=''
while getopts  'hFf:T:' option ; do
	case "$option" in
		h) showUsage ; exit 0 ;;
		f) FMT=$OPTARG ;;
		T) typeset -ft ${OPTARG//,/ } ;;
		F) typeset +f ; exit 0 ;;
	esac
done
X=$(( OPTIND - 1 ))
shift $X

[[ -z $1 ]] && showUsage && exit
while [[ -n "$1" ]]; do
	if [[ ! -r "$1" ]]; then
		print -u2 "Unable to read $1 - ignored"
	elif [[ ! -f "$1" ]]; then
		print -u2 "$1 is not a file - ignored."
	else
		FILES+="$1 "
	fi
	shift
done

for FILE in $FILES ; do
	analyze $FILE
done
printResults $FMT
