#!/bin/ksh93 

# License: CDDL 1.0
# Copyright (C) 2013 by Jens Elkner . All rights reserved.

TESTDIR='/pool1/test'
STATDIR='/tmp/stats'
FILEBENCH=${.sh.file%/*}/go_filebench
integer DDSZ=32
export LC_CTYPE=C LC_NUMERIC=C LC_MESSAGES=C LC_TIME=C LC_COLLATE=C LC_MEASUREMENT=C

# bogus:	compflow_demo
BENCHS='
	bringover
	copyfiles
	createfiles
	deletefiles
	fileserver
	listdirs
	makedirs
	mongo
	multistreamread
	multistreamreaddirect
	multistreamwrite
	multistreamwritedirect
	networkfs
	oltp
	openfiles
	randomfileaccess
	randomread
	randomrw
	randomwrite
	ratelimcopyfiles
	removedirs
	singlestreamread
	singlestreamreaddirect
	singlestreamwrite
	singlestreamwritedirect
	tpcso
	varmail
	videoserver
	webproxy
	webserver
'
BENCHS2='
	fileserver
	multistreamread
	multistreamwrite
	oltp
	singlestreamread
	singlestreamwrite
	varmail
	webproxy
'

function ddTest {
	typeset F=''
	typeset -a OUT=( )
	typeset -a -F1 RES=( )
	integer I OFFR=7  OFFW=1
	[[ ${ uname -s ; } == 'Linux' ]] && OFFR=16 && OFFW=1
	cd $TESTDIR/ || exit 1
	rm -rf * ; sync ; sleep 1
	print "Running sequential write test ..."
	typeset MKFILE=${ whence mkfile ; }
	if [[ -z ${MKFILE} ]]; then
		print -u2 "mkfile not found (you may use illumos src to get it). Skipping ddTest."
		return 1
	fi
	for (( I=0; I < 4; I++ )); do
		F+='bla'
		OUT=( ${ /usr/bin/time -p sh -c "${MKFILE} ${DDSZ}g $F ; sync" 2>&1 1>/dev/null; } )
		print "${I}:  ${OUT[@]}" >>$STATDIR/dd.err
		RES[$I]=$(( ( ${DDSZ} * 1024 ) / ${OUT[${OFFW}]} ))
	done
	print "write\t${RES[*]// /\\t}\tMiB/s" >$STATDIR/dd.lst
	chmod 644 b* ; sync ; sleep 1
	F=''
	typeset -a -F1 RES=( )
	print "Running sequential read test ..."
	for (( I=0; I < 4; I++ )); do
		F+='bla'
		OUT=( ${ /usr/bin/time -p dd if=$F of=/dev/null bs=128k 2>&1 1>/dev/null ; } )
		print "${I}:  ${OUT[@]}" >>$STATDIR/dd.err
		RES[$I]=$(( ( ${DDSZ} * 1024 ) / ${OUT[${OFFR}]} ))
	done
	print "read\t${RES[*]// /\\t}\tMiB/s" >>$STATDIR/dd.lst
}

function doBench {
	typeset DIR B
	integer I=$1
	if [[ ${ uname -s; } == 'Linux' ]]; then
		print 0 >/proc/sys/kernel/randomize_va_space
		# on Ubuntu default seem to be 32M
		print 268435456 >/proc/sys/kernel/shmmax
		FILEBENCH=/opt/filebench/bin/filebench
	fi
	for DIR in $TESTDIR $STATDIR ; do
		if [[ ! -d $DIR ]]; then
			mkdir -p $DIR || exit 1
		fi
	done
	cd $TESTDIR/ || exit 1

	(( ! $I  )) && ddTest
	for (( I=0; I < 4; I++ )); do
		print "Starting pass $I ..."
		# takes ~ 1h
		for B in $BENCHS2 ; do
			rm -rf * ; sync ; sleep 1
			print "Running $B ..."
			print -e "load $B\nset \$dir=$TESTDIR\nrun 60" | \
				$FILEBENCH >$STATDIR/${B}.${I}.out
		done
	done
}

function printHead {
	print 'FILEBENCHs:'
	print '==========='
	printf '    %=10s %=10s %=20s %=10s %=10s %=10s\n' \
		'ops' 'ops/s' 'r/w' 'mb/s' 'cpu/op' 'latency'
	printf '    %=10s %=10s %=20s %=10s %=10s %=10s' \
		' '   ' '     ' '   ' '    '[µs]'   '[ms]'
}

function analyze {
	(( ! $# )) && return
	typeset OPS OPS_PS CRW TP CPUOPS LAT
	print "\n${1%.+([0-9]).out}:"
	# fileserver.0.out:IO Summary:      7394799 ops, 123016.4 ops/s, (11183/22367 r/w) 2981.8mb/s,    677us cpu/op,   1.0ms latency
	# BUT fileserver 1.4.9.1 (linux) has to more fields infront:
	# fileserver.0.out:31983: 61.707: IO Summary: 372110 ops, 6201.294 ops/s, (564/1128 r/w), 147.4mb/s,   1502us cpu/op,   2.0ms latency
	grep 'IO Summary' "$@" | while read LINE ; do
		F=( ${LINE#*IO Summary:} )
		OPS=${F[0]}
		OPS_PS=${F[2]}
		CRW=${F[4]#\(}
		TP=${F[6]%mb*}
		CPUOPS=${F[7]%us}
		LAT=${F[9]%ms}
		printf '    %10s %10s %=20s %10s %10s %10s\n' \
			"$OPS" "$OPS_PS" "$CRW" "$TP" "$CPUOPS" "$LAT"
	done
}

function doCompare {
	typeset FNAME LAST='' FILES='' BN
	cd $STATDIR || exit 1

	print 'sequential (dd/mkfile):'
	print '======================='
	sed -e 's,^,	,' dd.lst
	print '\n'

	printHead
	ls -1 *.out | sort -u | while read FNAME ; do
		[[ ! -f $FNAME ]] && continue
		BN=${FNAME%+([0-9]).out}
		if [[ $BN != "$LAST" ]]; then
			analyze $FILES
			FILES="$FNAME"
			LAST=$BN
		else
			FILES+=" $FNAME"
		fi
	done
	analyze $FILES
}

function showUsage {
	print "Usage: ${.sh.file##*/}" '[-h] [-a] [-c] [-d] [-S size] \
                                    [-s statdir] [-t testdir]

  -h         .. print this help and exit.
  -a         .. If given, do all filebench benchmarks.
  -c         .. Compare/summarize the results in statdir, only.
  -d         .. Skip the sequential write/read test (mkfile, dd).
  -S size    .. Size of the file in GiB to use in the sequential w/r test.
                Default: 32
  -s statdir .. the directory with the output of all workload benchmarks.
                default: /tmp/stats
  -t testdir .. directory under test (used as argument for "set $dir=...").
                NOTE: Everything in this directory gets removed unconditionally!
                default: /pool1/test
'
}

integer COMPARE=0 SKIPDD=0 S=$DDSZ

while getopts  'FhcadT:t:s:S:' option ; do
	case "$option" in
		h) showUsage ; exit 0 ;;
		s) STATDIR="$OPTARG" ;;
		t) TESTDIR="$OPTARG" ;;
		c) COMPARE=1 ;;
		a) BENCHS2="$BENCH" ;;
		d) SKIPDD=1 ;;
		S) DDSZ=$OPTARG ; (( $DDSZ < 1 )) && DDSZ=$S ;;
		T) typeset -ft ${OPTARG//,/ } ;;
		F) typeset +f ; exit 0 ;;
	esac
done

(( COMPARE )) && doCompare || doBench $SKIPDD
